#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "../Misc/my_dbug.h"
#include "../Misc/mem_management.h" 
#include "../Syntaxtree/syntaxtree.h"
#include "../Typechecker/tc_util.h"
#include "ml_limits.h"
//#include "eval_global.h"
#include "eval_expand.h"
#include "eval_simplify.h"


/* Just called by evalThisNode
 * @return -1 = no further evaluation possible
 *          0 = are not equal
 *          1 = are equal
 */
static gint  
applyEqual(Expr_node_ptr c1, Expr_node_ptr c2)
{
  gint res;
  Expr_node_ptr act_elm1, act_elm2, act_arg1, act_arg2;

  DBUG_ENTER("applyEqual");
  DBUG_ASSERT(c1 != NULL);
  DBUG_ASSERT(c2 != NULL);
  DBUG_PRINT("eval", ("applyEqual"));

  if ( (c1->nodetype == e_constant) || (c1->nodetype == e_listkeeper)  
       || (c1->nodetype == e_list) || (c1->nodetype == e_constructor) ) {
    if (c1->nodetype == c2->nodetype) {
      switch (c1->nodetype) {
      case e_constant:
	res = areMLValuesEqual(c1, c2);
	break;

      case e_listkeeper:
	if ((LIST_EXPR_NODE(c1) == NULL) || (LIST_EXPR_NODE(c2) == NULL)) {
	  res = (LIST_EXPR_NODE(c1) == LIST_EXPR_NODE(c2));
	} else {
	  res = applyEqual(LIST_EXPR_NODE(c1), LIST_EXPR_NODE(c2));
	}
	break;

      case e_list:
	//check for different elements
	act_elm1 = c1;
	act_elm2 = c2;
	res = 1; //TRUE;
	while ((res == 1) && (act_elm1 != NULL) && (act_elm2 != NULL)) {
	  res = applyEqual(LISTITEM_EXPR_NODE(act_elm1), 
			   LISTITEM_EXPR_NODE(act_elm2));
	  act_elm1 = LISTNEXT_EXPR_NODE(act_elm1);
	  act_elm2 = LISTNEXT_EXPR_NODE(act_elm2);
	}
	//check for different length
	if ( (res == 1) && (!((act_elm1 == NULL) && (act_elm2 == NULL))) ) {
	  res = 0; //FALSE;
	}
	break;

      case e_constructor:
	//check for different constructors
	if (0 != strcmp(c1->strrep, c2->strrep)) {
	  res = 0; //FALSE;
	} else {
	  //check for different arguments
	  act_arg1 = ARGLIST_EXPR_NODE(c1);
	  act_arg2 = ARGLIST_EXPR_NODE(c2);
	  res = 1; //TRUE;
	  while ( (res == 1) && (act_arg1 != NULL) && (act_arg2 != NULL) ) {
	    res = applyEqual(ARG_EXPR_NODE(act_arg1), 
			     ARG_EXPR_NODE(act_arg2));
	    act_arg1 = NEXT_EXPR_NODE(act_arg1);
	    act_arg2 = NEXT_EXPR_NODE(act_arg2);
	  }
	  //typechecker guarantees same nr of arguments!!!
	  DBUG_ASSERT((res != 1) || ((act_arg1 == NULL)&&(act_arg2 == NULL)));
	}
      default:
	printf_expr_node(c1);
	res = -1; //only to avoid a compiler-warning
	DBUG_ASSERT(0);
      } //switch nodetype

    } else {
      //nodetypes of c1 and c2 are different
      DBUG_PRINT("eval", ("e_equal avoided"));
      res = -1;
    }
  } else {
    //c1->nodetype is wrong/not evaluatable
    DBUG_PRINT("eval", ("e_equal avoided"));
    res = -1;
  }

  DBUG_RETURN(res);
} //applyEqual


/* Just called by evalThisNode.
 * @return -1 no further evaluation possible
 */
static gint  
applyLength(Expr_node_ptr keeper)
{
  gint res;
  Expr_node_ptr c1;

  DBUG_ENTER("applyLength");
  DBUG_ASSERT(keeper != NULL);
  DBUG_PRINT("eval", ("applyLength"));

  c1 = LIST_EXPR_NODE(keeper);
  if (keeper->nodetype != e_listkeeper) {
    res = -1;
  } else {
    if (c1 == NULL) {
      res = 0;
    } else {
      if (c1->nodetype != e_list) {
	res = -1;
      } else {
	res = 0;
	while ((res >= 0) && (c1 != NULL)) {
	  res++;
          if (c1->nodetype != e_list) {
	    res = -1;
	  }
	  c1 = LISTNEXT_EXPR_NODE(c1);
	} //while
      }
    }
  }

  DBUG_RETURN(res);
} //applyLength


/* @return -1 in case of an runtimeerror
 *          0 if further evaluation is not possible
 *          1 in case of a successful application
 */
gint 
evalThisNode(Expr_node_ptr parent_of_node, Expr_node_ptr node, 
	     EvalMode mode, gboolean simplify_only)
{
  gint res = 0;
  gint err, len;
  gdouble tmp_double;
  Expr_node_ptr c1, c2, c3, tmp;
  struct _Mlval_struct val;

  
  DBUG_ENTER("evalThisNode");
  DBUG_ASSERT(node != NULL);
  DBUG_PRINT("eval", ("evalThisNode '%s'", 
		      (node->strrep != NULL) ? node->strrep : "<no strrep>"));

  eval_stats.expansiondepth = node->expansiondepth;
  switch (node->nodetype) {
  case e_errornode:
    runtime_error("Expression contains an ERRORNODE");
    res = -1;
    break;

  case e_arg_list:
  case e_listkeeper:
  case e_list:
  case e_constant:
  case e_nil_const:
  case e_product:    
    
    case e_letvars_list:
    res = 0;
    break;

  case e_unary_not:
    c1 = OPERAND_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == bool_type_index);
      val.val    = c1->value;
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);

      val.val.mlbool = !val.val.mlbool;

      tmp = create_expr_constant(val, c1->valuetype);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_unary_minus:
    c1 = OPERAND_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      val.val    = c1->value;
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);

      if (c1->valuetype == int_type_index) {
	tmp_double = -val.val.mlint;
	if (isIntBetweenLimits(tmp_double)) {
	  val.val.mlint = (gint)tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	} else {
	  tmp = create_errornode();
	  res = -1;
	}
      } else {
	DBUG_ASSERT(c1->valuetype == real_type_index);
	tmp_double = -val.val.mlreal;
	if (isRealBetweenLimits(tmp_double)) {
	  val.val.mlreal = tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	} else {
	  tmp = create_errornode();
	  res = -1;
	}
      }

      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_andalso:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == bool_type_index);
      DBUG_ASSERT(c2->valuetype == bool_type_index);
      if (c1->value.mlbool == FALSE) {
	val.lineno     = LINENO_NODE(node);
	val.colno      = COLNO_NODE(node);
	val.val.mlbool = FALSE;
	tmp = create_expr_constant(val, bool_type_index);
      } else {
	tmp = c2;
	//avoid that right operand will be freed when replacing subtree
	RIGHT_OPERAND_EXPR_NODE(node) = NULL;
      } //if value == FALSE
      replaceSubtree(parent_of_node, node, tmp);
      node = NULL; //avoid trouble
      eval_stats.nr_evaluations++;
      res = 1;
    } //if e_constant
    break;

  case e_orelse:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == bool_type_index);
      DBUG_ASSERT(c2->valuetype == bool_type_index);
      if (c1->value.mlbool == TRUE) {
	val.lineno     = LINENO_NODE(node);
	val.colno      = COLNO_NODE(node);
	val.val.mlbool = TRUE;
	tmp = create_expr_constant(val, bool_type_index);
      } else {
	tmp = c2;
	//avoid that right operand will be freed when replacing subtree
	RIGHT_OPERAND_EXPR_NODE(node) = NULL;
      } //if value == TRUE
      replaceSubtree(parent_of_node, node, tmp);
      node = NULL; //avoid trouble
      eval_stats.nr_evaluations++;
      res = 1;
    } //if e_constant
    break;

  case e_equal:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    err = applyEqual(c1, c2);
    if (err != -1) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlbool = (err == 1/*TRUE*/);

      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_unequal:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    err = applyEqual(c1, c2);
    if (err != -1) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlbool = (err == 0/*FALSE*/);

      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_lt:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	val.val.mlbool = (c1->value.mlint < c2->value.mlint);
      } else if (c1->valuetype == real_type_index) {
	val.val.mlbool = (c1->value.mlreal < c2->value.mlreal);
      } else {
	DBUG_ASSERT(c1->valuetype == string_type_index);
	val.val.mlbool = (0 < strcmp(c1->value.mlstring, c2->value.mlstring));
      }
      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_le:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	val.val.mlbool = (c1->value.mlint <= c2->value.mlint);
      } else if (c1->valuetype == real_type_index) {
	val.val.mlbool = (c1->value.mlreal <= c2->value.mlreal);
      } else {
	DBUG_ASSERT(c1->valuetype == string_type_index);
	val.val.mlbool = (0 <= strcmp(c1->value.mlstring, c2->value.mlstring));
      }
      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_gt:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	val.val.mlbool = (c1->value.mlint > c2->value.mlint);
      } else if (c1->valuetype == real_type_index) {
	val.val.mlbool = (c1->value.mlreal > c2->value.mlreal);
      } else {
	DBUG_ASSERT(c1->valuetype == string_type_index);
	val.val.mlbool = (0 > strcmp(c1->value.mlstring, c2->value.mlstring));
      }
      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_ge:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	val.val.mlbool = (c1->value.mlint >= c2->value.mlint);
      } else if (c1->valuetype == real_type_index) {
	val.val.mlbool = (c1->value.mlreal >= c2->value.mlreal);
      } else {
	DBUG_ASSERT(c1->valuetype == string_type_index);
	val.val.mlbool = (0 >= strcmp(c1->value.mlstring, c2->value.mlstring));
      }
      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_plus:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	tmp_double = ((gdouble)c1->value.mlint) + ((gdouble)c2->value.mlint);
	if (isIntBetweenLimits(tmp_double)) {
	  val.val.mlint = (gint)tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      } else {
	DBUG_ASSERT(c1->valuetype == real_type_index);
	tmp_double = c1->value.mlreal + c2->value.mlreal;
	if (isRealBetweenLimits(tmp_double)) {
	  val.val.mlreal = tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      }
      replaceSubtree(parent_of_node, node, tmp);
    } else DBUG_PRINT("eval", ("e_plus avoided"));
    break;

  case e_minus:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	tmp_double = ((gdouble)c1->value.mlint) - ((gdouble)c2->value.mlint);
	if (isIntBetweenLimits(tmp_double)) {
	  val.val.mlint = (gint)tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      } else {
	DBUG_ASSERT(c1->valuetype == real_type_index);
	tmp_double = c1->value.mlreal - c2->value.mlreal;
	if (isRealBetweenLimits(tmp_double)) {
	  val.val.mlreal = tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_mult:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	tmp_double = ((gdouble)c1->value.mlint) * ((gdouble)c2->value.mlint);
	if (isIntBetweenLimits(tmp_double)) {
	  val.val.mlint = (gint)tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      } else {
	DBUG_ASSERT(c1->valuetype == real_type_index);
	tmp_double = c1->value.mlreal * c2->value.mlreal;
	if (isRealBetweenLimits(tmp_double)) {
	  val.val.mlreal = tmp_double;
	  tmp = create_expr_constant(val, c1->valuetype);
	  eval_stats.nr_evaluations++;
	  res = 1;
	}else {
	  tmp = create_errornode();
	  res = -1;
	}
      }
      replaceSubtree(parent_of_node, node, tmp);
    } else DBUG_PRINT("eval", ("e_mult avoided"));
    break;

  case e_mod:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == int_type_index);
      DBUG_ASSERT(c2->valuetype == int_type_index);
      if (c2->value.mlint != 0) {
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val.mlint = c1->value.mlint % c2->value.mlint;

	tmp = create_expr_constant(val, c1->valuetype);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Integer-division by zero");
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_int_div:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == int_type_index);
      DBUG_ASSERT(c2->valuetype == int_type_index);
      if (c2->value.mlint != 0) {
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val.mlint = (gint) (c1->value.mlint / c2->value.mlint);

	tmp = create_expr_constant(val, c1->valuetype);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Integer-division by zero");
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_real_div:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      DBUG_ASSERT(c2->valuetype == real_type_index);
      if (c2->value.mlreal != 0.0) {
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val.mlreal = c1->value.mlreal / c2->value.mlreal;

	tmp = create_expr_constant(val, c1->valuetype);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Real-division by zero");
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_string_concat:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == string_type_index);
      DBUG_ASSERT(c2->valuetype == string_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlstring = mtstrconcat(2, c1->value.mlstring, 
				        c2->value.mlstring);

      tmp = create_expr_constant(val, string_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_prepend:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ( ( (c1->nodetype == e_constant) || (c1->nodetype == e_listkeeper)  
	   || (c1->nodetype == e_constructor)
	 ) && (c2->nodetype == e_listkeeper) ) {

      c3 = create_expr_list(c1, LIST_EXPR_NODE(c2));
      c3->valuetype = node->valuetype;
      LINENO_NODE(c3) = LINENO_NODE(node);
      COLNO_NODE(c3)  = COLNO_NODE(node);

      LIST_EXPR_NODE(c2) = c3;
      tmp = c2;

      //avoid that head and tail will be freed when replacing subtree
      LEFT_OPERAND_EXPR_NODE(node)  = NULL;
      RIGHT_OPERAND_EXPR_NODE(node) = NULL;
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_list_concat:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_listkeeper) && (c2->nodetype == e_listkeeper)) {
      //is left list empty?
      if (LIST_EXPR_NODE(c1) == NULL) {
	//avoid that right list will be freed when replacing subtree
	RIGHT_OPERAND_EXPR_NODE(node) = NULL;
	replaceSubtree(parent_of_node, node, c2);
      } else {
	//seek last element of left list
	tmp = LIST_EXPR_NODE(c1);
	while (NEXT_EXPR_NODE(tmp) != NULL) {
	  tmp = NEXT_EXPR_NODE(tmp);
	}
	//concat lists
	NEXT_EXPR_NODE(tmp) = LIST_EXPR_NODE(c2);
	//avoid that list(s) will be freed when replacing subtree
	LEFT_OPERAND_EXPR_NODE(node)  = NULL;
	LIST_EXPR_NODE(c2)            = NULL;
	replaceSubtree(parent_of_node, node, c1);
      }
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_abs:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if        (c1->valuetype == int_type_index) {
	val.val.mlint = (c1->value.mlint >= 0 ?  c1->value.mlint 
			                      : -c1->value.mlint);
      } else {
	DBUG_ASSERT(c1->valuetype == real_type_index);
	val.val.mlreal = (c1->value.mlreal >= 0 ?  c1->value.mlreal
                                                : -c1->value.mlreal);
      }
      tmp = create_expr_constant(val, c1->valuetype);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_min:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == int_type_index);
      DBUG_ASSERT(c2->valuetype == int_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlint = (c1->value.mlint < c2->value.mlint 
	               ? c1->value.mlint : c2->value.mlint);
      tmp = create_expr_constant(val, int_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_max:
    c1 = LEFT_OPERAND_EXPR_NODE(node);
    c2 = RIGHT_OPERAND_EXPR_NODE(node);
    if ((c1->nodetype == e_constant) && (c2->nodetype == e_constant)) {
      DBUG_ASSERT(c1->valuetype == int_type_index);
      DBUG_ASSERT(c2->valuetype == int_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlint = (c1->value.mlint > c2->value.mlint 
	               ? c1->value.mlint : c2->value.mlint);
      tmp = create_expr_constant(val, int_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_sqrt:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      if (c1->value.mlreal > 0) { 
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val.mlreal = sqrt(c1->value.mlreal);

	tmp = create_expr_constant(val, real_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Squareroot of negative value");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_sin:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlreal = sin(c1->value.mlreal);

      tmp = create_expr_constant(val, real_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_cos:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlreal = cos(c1->value.mlreal);

      tmp = create_expr_constant(val, real_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_exp:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      tmp_double = exp(c1->value.mlreal);
      if (isRealBetweenLimits(tmp_double)) {
	val.val.mlreal = tmp_double;
	tmp = create_expr_constant(val, real_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	tmp = create_errornode();
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_ln:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      if (c1->value.mlreal > 0) { 
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	tmp_double = log(c1->value.mlreal);
	if (isRealBetweenLimits(tmp_double)) {
	  val.val.mlreal = tmp_double;
	  tmp = create_expr_constant(val, real_type_index);
	  eval_stats.nr_evaluations++;
	  res = 1;
	} else {
	  tmp = create_errornode();
	  res = -1;
	}
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Logarithmus of a value lower or equal to zero");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_real:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == int_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlreal = (gdouble)c1->value.mlint;

      tmp = create_expr_constant(val, real_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_truncate:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == real_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      tmp_double = ((gint)c1->value.mlreal);
      if (isIntBetweenLimits(tmp_double)) {
	val.val.mlint = (gint)tmp_double;
	tmp = create_expr_constant(val, int_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	tmp = create_errornode();
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_size:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == string_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlint = strlen(c1->value.mlstring);

      tmp = create_expr_constant(val, int_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_ord:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == string_type_index);
      if (strlen(c1->value.mlstring) > 0) { 
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val.mlint = (gint)(c1->value.mlstring[0]);

	tmp = create_expr_constant(val, int_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Ord of an empty string");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_chr:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      gchar single_char[2];
      DBUG_ASSERT(c1->valuetype == int_type_index);
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      if ((0 <= c1->value.mlint) && (c1->value.mlint <= 255)) {
	single_char[0]   = (gchar)(c1->value.mlint);
	single_char[1]   = '\0';
	val.val.mlstring = mtstrdup(single_char);

	tmp = create_expr_constant(val, string_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	tmp = create_errornode();
	runtime_error("chr-argument not between 0 and 255");
	res = -1;
      }
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_substring:
    c1 = COND_EXPR_NODE(node);
    c2 = THEN_EXPR_EXPR_NODE(node);
    c3 = ELSE_EXPR_EXPR_NODE(node);
    if ( (c1->nodetype == e_constant) && (c2->nodetype == e_constant)
	 && (c3->nodetype == e_constant) ) {
      DBUG_ASSERT(c1->valuetype == string_type_index);
      DBUG_ASSERT(c2->valuetype == int_type_index);
      DBUG_ASSERT(c3->valuetype == int_type_index);

      len = strlen(c1->value.mlstring);
      if ( (c2->value.mlint >= 0) && (c3->value.mlint >= 0) 
	   && (c2->value.mlint + c3->value.mlint <= len) ) {
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	c1->value.mlstring += c2->value.mlint;
	c1->value.mlstring[c3->value.mlint] = '\0';
	val.val.mlstring = mtstrdup(c1->value.mlstring);

	tmp = create_expr_constant(val, string_type_index);
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Substring - wrong parameters");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_null:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_listkeeper) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlbool = (LIST_EXPR_NODE(c1) == NULL);

      tmp = create_expr_constant(val, bool_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_hd:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_listkeeper) {
      if (LIST_EXPR_NODE(c1) != NULL) {
	/*
	val.lineno = LINENO_NODE(node);
	val.colno  = COLNO_NODE(node);
	val.val = LISTITEM_EXPR_NODE(LIST_EXPR_NODE(c1))->value;
	tmp = create_expr_constant(val, 
	        LISTITEM_EXPR_NODE(LIST_EXPR_NODE(c1))->valuetype);
	*/
	tmp = LISTITEM_EXPR_NODE(LIST_EXPR_NODE(c1));
	//avoid that the list-tail will be freed when replacing subtree
	LISTITEM_EXPR_NODE(LIST_EXPR_NODE(c1)) = NULL;
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Head of an empty list");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_tl:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_listkeeper) {
      if (LIST_EXPR_NODE(c1) != NULL) {
	c3 = LISTNEXT_EXPR_NODE(LIST_EXPR_NODE(c1));
	//avoid that the list-tail will be freed when freeing the list-head
	LISTNEXT_EXPR_NODE(LIST_EXPR_NODE(c1)) = NULL;
	free_expr_node(LIST_EXPR_NODE(c1));
	LIST_EXPR_NODE(c1) = c3;
        tmp = c1;
	//avoid that the resulting list will be freed when replacing subtree
        ARG_EXPR_NODE(node) = NULL;
	eval_stats.nr_evaluations++;
	res = 1;
      } else {
	//handle the runtimeerror
	tmp = create_errornode();
	runtime_error("Tail of an empty list");
	res = -1;
      }  
      replaceSubtree(parent_of_node, node, tmp);
    }
    break;

  case e_length:
    c1 = ARG_EXPR_NODE(node);
    len = applyLength(c1);
    if (len >= 0) {
      val.lineno = LINENO_NODE(node);
      val.colno  = COLNO_NODE(node);
      val.val.mlint = len;

      tmp = create_expr_constant(val, int_type_index);
      replaceSubtree(parent_of_node, node, tmp);
      eval_stats.nr_evaluations++;
      res = 1;
    }
    break;

  case e_rev:
    c1 = ARG_EXPR_NODE(node);
    if (c1->nodetype == e_listkeeper) {
      gint len;

      len = applyLength(c1);
      if (len != -1) {
	if (len <= 1) {
	  tmp = c1;
	  //avoid that list will be freed when replacing subtree
	  ARG_EXPR_NODE(node) = NULL;
	} else {
	  Expr_node_ptr prev, act, next;//, new_last;

	  prev = NULL;
	  act  = LIST_EXPR_NODE(c1);
	  next = LISTNEXT_EXPR_NODE(act);
	  while (next != NULL) {
	    LISTNEXT_EXPR_NODE(act) = prev;
	    prev = act;
	    act  = next;
	    next = LISTNEXT_EXPR_NODE(next);
	  }
	  LISTNEXT_EXPR_NODE(act) = prev;
	  LIST_EXPR_NODE(c1) = act;
	  //avoid that list will be freed when replacing subtree
	  ARG_EXPR_NODE(node) = NULL;
	  tmp = c1;
	} //if len <= 1

	replaceSubtree(parent_of_node, node, tmp);
	eval_stats.nr_evaluations++;
	res = 1;
      } //if len != -1
    }
    break;

  case e_project:
  { 
    int pos = LEFT_OPERAND_EXPR_NODE(node)->value.mlint;
    if ( RIGHT_OPERAND_EXPR_NODE(node)->nodetype == e_product ) {
      Expr_node_ptr part = ARG_EXPR_NODE(RIGHT_OPERAND_EXPR_NODE(node));
      while ( part != NULL && pos > 0 ) {
        part = NEXT_EXPR_NODE(part);;
        pos--;
      }
      if ( part != NULL && pos == 0 ) {
        Expr_node_ptr found = ARG_EXPR_NODE(part);
        ARG_EXPR_NODE(part) = NULL;
        replaceSubtree(parent_of_node, node, found);
        eval_stats.nr_evaluations++;
      }
    }
  }
    break;

  case e_let:
   
    if ( !simplify_only ) {
    
    Expr_node_ptr vars = LET_VARS_EXPR_NODE(node);
    Expr_node_ptr valr = LET_VAL_EXPR_NODE(node);
    Expr_node_ptr expr = LET_EXPR_EXPR_NODE(node);
      
    // replace variables in expr
    
    // count number of vars
    int varcount = 0;
    Expr_node_ptr varlist = vars; 
    while ( varlist != NULL ) {
      varlist = NEXT_EXPR_NODE(varlist); 
      varcount++;
    }

    
    if ( varcount==1 || valr->nodetype == e_product ) {
      
    // iterate over all variables
    Expr_node_ptr varlist = vars;
    Expr_node_ptr byExpr;

    int varpos = 0;
    while ( varlist != NULL ) {
      Expr_node_ptr var = ARG_EXPR_NODE(varlist); // var points to variable
      
      // find value
      int valpos = 0;
      Expr_node_ptr val_runner = LISTITEM_EXPR_NODE(valr);
      while ( val_runner != NULL && valpos < varpos ) {
        val_runner = LISTNEXT_EXPR_NODE(val_runner);
        valpos++;
      }
      byExpr = val_runner != NULL && ARG_EXPR_NODE(val_runner) != NULL ? ARG_EXPR_NODE(val_runner) : valr;

      if ( varcount == 1 && valr->nodetype != e_product ) {
        byExpr = valr;
      }
      
      // replace occs
      if ( byExpr != NULL ) {
        replaceAllOccurencesAndSetExpansiondepth(
          var->strrep, // varname
          node, // parent of Expr
          expr, // inExpr
          byExpr , // byExpr
          eval_stats.expansiondepth //depth
         ); 
      }

      // next
      varlist = NEXT_EXPR_NODE(varlist); 
      varpos++;
    }

    //avoid that expr will be freed when replacing subtree
    expr = LET_EXPR_EXPR_NODE(node);
    LET_EXPR_EXPR_NODE(node) = NULL;

    replaceSubtree(parent_of_node, node, expr);

    node = NULL; //avoid trouble
    eval_stats.nr_evaluations++;
    res = 1;
    
    // if it's not yet a product that we can easily divide
    } else {
      // we need to use the projections here
      
      // iterate over all variables
      Expr_node_ptr varlist = vars;
      Expr_node_ptr posNode;
      Expr_node_ptr newNode;
      int varpos = 0;
      while ( varlist != NULL ) {
        Expr_node_ptr var = ARG_EXPR_NODE(varlist); // var points to variable
        
        // make a copy of the val-tree
        Expr_node_ptr copyOfVal = duplicateSubtree(valr);
        
        // create position
        val.lineno     = LINENO_NODE(node);
        val.colno      = COLNO_NODE(node);
        val.val.mlint = varpos;
        posNode = create_expr_constant(val, int_type_index);
        
        // create projection
        newNode = create_expr_project(posNode, copyOfVal);
        
        replaceAllOccurencesAndSetExpansiondepth(
            var->strrep, // varname
        node, // parent of Expr
        expr, // inExpr
        newNode, // byExpr
        eval_stats.expansiondepth //depth
                                                ); 

        // next
        varlist = NEXT_EXPR_NODE(varlist); 
        varpos++;
      }
         //avoid that expr will be freed when replacing subtree
      expr = LET_EXPR_EXPR_NODE(node);
      LET_EXPR_EXPR_NODE(node) = NULL;

      replaceSubtree(parent_of_node, node, expr);

      node = NULL; //avoid trouble
      eval_stats.nr_evaluations++;
      res = 1;

     }
    }
    break;
    
    case e_variable:
      break;
  case e_ifthenelse:
    c1 = COND_EXPR_NODE(node);
    if (c1->nodetype == e_constant) {
      DBUG_ASSERT(c1->valuetype == bool_type_index);
      if (c1->value.mlbool == FALSE) {
	tmp = ELSE_EXPR_EXPR_NODE(node);
	//avoid that else-case will be freed when replacing subtree
	ELSE_EXPR_EXPR_NODE(node) = NULL;
      } else {
	tmp = THEN_EXPR_EXPR_NODE(node);
	//avoid that then-case will be freed when replacing subtree
	THEN_EXPR_EXPR_NODE(node) = NULL;
      }
      replaceSubtree(parent_of_node, node, tmp);
      node = NULL; //avoid trouble
      eval_stats.nr_evaluations++;
      res = 1;
    } //if e_constant
    break;

  case e_constructor:
    res = 0;
    break;

  case e_userfun:
    if (!simplify_only) {
      if (expandThisFunctioncall(parent_of_node, node) == TRUE) {
	res = 1;
      } else {
	res = -1;
      } //if-else
    } //if
    break;

  default:
    printf_expr_node(node);
    res = -1; //only to avoid a compiler-warning
    DBUG_ASSERT(0);
  } //switch

  { static int count = 0;
  if ((count >= 10) && (fpUpdateStatistics != NULL)) {
    (*fpUpdateStatistics)();
    count = 0;
  } else {
    count++;
  } //if
  } //block

  DBUG_PRINT("eval", ("res = %d", res));
  DBUG_RETURN(res);
} //evalThisNode


/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
