#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../Misc/my_dbug.h"
#include "../Misc/utilities.h"
#include "../Misc/mem_management.h"
#include "syntaxtree.h"


/* a local typedef */
typedef enum {e_unused = -1, e_no = 0, e_yes = 1} TIsFirstElmInList;

/* forward declarations */
void printf_treeparts(GList* tree_parts);


GList* syntax_tree = NULL;
/* In case of an error while building up the syntaxtree there allready exist 
 * seperated parts of the syntaxtree, which can't be freed. To become able to 
 * free these parts too, each root of these parts is stored in the 
 * 'tree_parts'. If some parts are put together by a new root-node, the roots 
 * of the parts are removed and the new root is added. A root is removed
 * completely, if a whole fun_node is parsed.
 */
GList* tree_parts = NULL;
gboolean keep_tree_parts_in_mind = FALSE; /* default is FALSE; should be set 
					   * to TRUE in scanner&parser */
void (*fpFree_expr_node_misc) (gpointer pMisc) = NULL; 
  /* Pointer to a function which frees the memory behind ->misc. 
   * Has to be able to handle a NULL-pointer as argument! */

/** Removes an element from the GList and frees the GList-struct, but 
  * does not free the data.
  * @return New first element of GList.
  */
GList*
myGlistRemove(GList* list, gpointer data)
{
  GList* tmp;

  DBUG_ENTER("myGlistRemove");

  tmp  = g_list_find(list, data);
  if (tmp != NULL) {
    list = g_list_remove_link(list, tmp);
    tmp->data = NULL; //avoid freeing the data (this has to be done elsewhere)
    g_list_free_1(tmp);
  } else {
    DBUG_PRINT( "seek", ("g_list_find(%p) failed", data) );
  }

  DBUG_RETURN(list);
} //myGlistRemove


/* @return a constant string representing the nodetype
 */
gchar* 
nodetype2str(Nodetype nt)
{
  gchar* strrep;

  DBUG_ENTER("nodetype2str");

  switch (nt) {
  case e_invalid      : strrep = "e_invalid"; break;
  case e_errornode    : strrep = "e_errornode"; break;
  case e_param_list   : strrep = "e_param_list"; break;
  case e_arg_list     : strrep = "e_arg_list"; break;
  case e_listkeeper   : strrep = "e_listkeeper"; break;
  case e_list         : strrep = "e_list"; break;
  case e_constant     : strrep = "e_constant"; break;
  case e_nil_const    : strrep = "e_nil_const"; break;
  case e_variable     : strrep = "e_variable"; break;
  case e_unary_not    : strrep = "e_unary_not"; break;
  case e_unary_minus  : strrep = "e_unary_minus"; break;
  case e_andalso      : strrep = "e_andalso"; break;
  case e_orelse       : strrep = "e_orelse"; break;
  case e_equal        : strrep = "e_equal"; break;
  case e_unequal      : strrep = "e_unequal"; break;
  case e_lt           : strrep = "e_lt"; break;
  case e_le           : strrep = "e_le"; break;
  case e_gt           : strrep = "e_gt"; break;
  case e_ge           : strrep = "e_ge"; break;
  case e_plus         : strrep = "e_plus"; break;
  case e_minus        : strrep = "e_minus"; break;
  case e_mult         : strrep = "e_mult"; break;
  case e_mod          : strrep = "e_mod"; break;
  case e_int_div      : strrep = "e_int_div"; break;
  case e_real_div     : strrep = "e_real_div"; break;
  case e_string_concat: strrep = "e_string_concat"; break;
  case e_prepend      : strrep = "e_prepend"; break;
  case e_list_concat  : strrep = "e_list_concat"; break;
  case e_abs          : strrep = "e_abs"; break;
  case e_min          : strrep = "e_min"; break;
  case e_max          : strrep = "e_max"; break;
  case e_sqrt         : strrep = "e_sqrt"; break;
  case e_sin          : strrep = "e_sin"; break;
  case e_cos          : strrep = "e_cos"; break;
  case e_exp          : strrep = "e_exp"; break;
  case e_ln           : strrep = "e_ln"; break;
  case e_real         : strrep = "e_real"; break;
  case e_truncate     : strrep = "e_truncate"; break;
  case e_size         : strrep = "e_size"; break;
  case e_ord          : strrep = "e_ord"; break;
  case e_chr          : strrep = "e_chr"; break;
  case e_substring    : strrep = "e_substring"; break;
  case e_null         : strrep = "e_null"; break;
  case e_hd           : strrep = "e_hd"; break;
  case e_tl           : strrep = "e_tl"; break;
  case e_length       : strrep = "e_length"; break;
  case e_rev          : strrep = "e_rev"; break;
  case e_ifthenelse   : strrep = "e_ifthenelse"; break;
  case e_constructor  : strrep = "e_constructor"; break;
  case e_userfun      : strrep = "e_userfun"; break;
  case e_wildcard     : strrep = "e_wildcard"; break;
  case e_product      : strrep = "e_product"; break;
  case e_let          : strrep = "e_let"; break;
  case e_letvars_list : strrep = "e_letvars_list"; break;
  case e_project      : strrep = "e_project"; break;
  default             : strrep = "<UNKNOWN>";
  } //switch

  DBUG_RETURN(strrep);
} //nodetype2str


static gchar* 
mlConstant2String(Expr_node_ptr pExpr)
{
  const int max_str = 256;
  gchar* res = NULL;
  gchar valstr[max_str];

  DBUG_ENTER("mlConstant2String");
  DBUG_ASSERT(pExpr != NULL);
  DBUG_ASSERT(pExpr->nodetype == e_constant);

  if        (pExpr->valuetype == 1) {
    res = mtstrdup("()");

  } else if (pExpr->valuetype == 2) {
    sprintf(valstr, "%s", (pExpr->value.mlint == 0 ? "false" : "true"));
    res = mtstrdup(valstr);

  } else if (pExpr->valuetype == 3) {
    sprintf(valstr, "%i", pExpr->value.mlint);
    res = mtstrdup(valstr);

  } else if (pExpr->valuetype == 4) {
    sprintf(valstr, "%e", pExpr->value.mlreal);
    strreplacechar(valstr, '-', '~');
    res = mtstrdup(valstr);

  } else if (pExpr->valuetype == 5) {
    res = mtstrconcat(3, "\"", pExpr->value.mlstring, "\"");

  } else {
    res = mtstrdup("?");
  }

  DBUG_RETURN(res);
} //mlConstant2String


static gchar*
infixop2strrep(Nodetype nt)
{
  gchar* strrep;

  DBUG_ENTER("infixop2strrep");

  switch (nt) {
  case e_andalso      : strrep = "andalso"; break;
  case e_orelse       : strrep = "orelse"; break;
  case e_equal        : strrep = "="; break;
  case e_unequal      : strrep = "<>"; break;
  case e_lt           : strrep = "<"; break;
  case e_le           : strrep = "<="; break;
  case e_gt           : strrep = ">"; break;
  case e_ge           : strrep = ">="; break;
  case e_plus         : strrep = "+"; break;
  case e_minus        : strrep = "-"; break;
  case e_mult         : strrep = "*"; break;
  case e_mod          : strrep = "mod"; break;
  case e_int_div      : strrep = "div"; break;
  case e_real_div     : strrep = "/"; break;
  case e_string_concat: strrep = "^"; break;
  case e_prepend      : strrep = "::"; break;
  case e_list_concat  : strrep = "@"; break;
  default             : strrep = "?"; DBUG_ASSERT(0);
  } //switch

  DBUG_RETURN(mtstrdup(strrep));
} //infixop2strrep


/* @return a constant string representing the basic functionname
 */
static gchar*
basicfun2strrep(Nodetype nt)
{
  gchar* strrep;

  DBUG_ENTER("basicfun2strrep");

  switch (nt) {
  case e_abs          : strrep = "abs"; break;
  case e_min          : strrep = "min"; break;
  case e_max          : strrep = "max"; break;
  case e_sqrt         : strrep = "sqrt"; break;
  case e_sin          : strrep = "sin"; break;
  case e_cos          : strrep = "cos"; break;
  case e_exp          : strrep = "exp"; break;
  case e_ln           : strrep = "ln"; break;
  case e_real         : strrep = "real"; break;
  case e_truncate     : strrep = "truncate"; break;
  case e_size         : strrep = "size"; break;
  case e_ord          : strrep = "ord"; break;
  case e_chr          : strrep = "chr"; break;
  case e_substring    : strrep = "substring"; break;
  case e_null         : strrep = "null"; break;
  case e_hd           : strrep = "hd"; break;
  case e_tl           : strrep = "tl"; break;
  case e_length       : strrep = "length"; break;
  case e_rev          : strrep = "rev"; break;
  default             : strrep = "?"; DBUG_ASSERT(0);
  } //switch

  DBUG_RETURN(mtstrdup(strrep));
} //basicfun2strrep


static void
printf_nodetype(Nodetype nt)
{
  DBUG_ENTER("printf_nodetype");

  printf("nodetype: %s (%i)", nodetype2str(nt), nt);

  DBUG_VOID_RETURN;
} //printf_nodetype


void
printf_value(MLType valuetype, MLValue value)
{
  gchar* str;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("printf_value");

  switch (valuetype) {
  case 1: printf("- value:  ():unit\n");  break;
  case 2: printf("- value:  %d:bool\n",     value.mlbool);  break;
  case 3: printf("- value:  %d:int\n",      value.mlint);  break;
  case 4: printf("- value:  %f:real\n",     value.mlreal);  break;
  case 5: printf("- value:  '%s':string\n", value.mlstring);  break;
  case 6: printf("- value:  nil:'a list\n");  break;
  default: 
    pSte = seekSymbolTypeindex(symbol_table, valuetype);
    str = (pSte == NULL ? "?" : pSte->id);
    printf("- value: not a basic type or a list\n");
    printf("- valuetype: %s (%i)\n",str , valuetype);
  } //switch

  DBUG_VOID_RETURN;
} //printf_value


void
printf_expr_node(Expr_node_ptr node)
{
  int i;

  DBUG_ENTER("printf_expr_node");

  printf("Expr_node_ptr: %p\tchildren: ", node);
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    printf("%p   ", node->child[i]);
  }
  printf("\n- ");
  printf_nodetype(node->nodetype);
  if (node->strrep == NULL) {
    printf("\n- node->strrep is NULL\n");
  } else {
    printf("\n- strrep  : '%s'\n",node->strrep);
  }
  printf_value(node->valuetype, node->value);

  DBUG_VOID_RETURN;
} //printf_expr_node



/** These create_expr_X functions allocate memory and set the node->nodetype 
  * and the used children, and all the rest to defaults (0 or NULL).
  * @return a pointer to the fresh allocated Expr_node_struct.
  */
Expr_node_ptr
create_dummy(void)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_dummy");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_invalid;
  res->strrep             = NULL;
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  DBUG_RETURN(res);
} //create_dummy


Expr_node_ptr
create_errornode(void)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_errornode");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_errornode;
  res->strrep             = mtstrdup("ERROR");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_errornode


/** These create_expr_X functions allocate memory and set the node->nodetype 
  * and the used children, and all the rest to defaults (0 or NULL).
  * @return a pointer to the fresh allocated Expr_node_struct.
  */
Expr_node_ptr
create_expr_param_list(Expr_node_ptr arg, Expr_node_ptr next)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_param_list");
  DBUG_ASSERT(arg != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_param_list;
  res->strrep             = NULL;
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARG_EXPR_NODE(res)  = arg;
  NEXT_EXPR_NODE(res) = next;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arg);
    tree_parts = myGlistRemove(tree_parts, next);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_param_list

Expr_node_ptr
    create_expr_letvars_list(Expr_node_ptr arg, Expr_node_ptr next)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_letvars_list");
  DBUG_ASSERT(arg != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_letvars_list;
  res->strrep             = mtstrdup("letvars_list");
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  arg->subnodetype        = e_boxed;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARG_EXPR_NODE(res)  = arg;
  NEXT_EXPR_NODE(res) = next;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arg);
    tree_parts = myGlistRemove(tree_parts, next);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_param_list

Expr_node_ptr
    create_expr_let(Expr_node_ptr vars, Expr_node_ptr val, Expr_node_ptr expr)
{
  Expr_node_ptr res;
  Expr_node_ptr varlist;
  int i;

  DBUG_ENTER("create_expr_let");
  DBUG_ASSERT(vars != NULL);
  DBUG_ASSERT(val != NULL);
  DBUG_ASSERT(expr != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_let;
  res->strrep             = mtstrdup("let");
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  res->child[0]  = vars;
  res->child[1]  = val;
  res->child[2]  = expr;
  
  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, vars);
    tree_parts = myGlistRemove(tree_parts, val);
    tree_parts = myGlistRemove(tree_parts, expr);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  
  
  // mark all variables
  varlist = vars;
  while ( varlist != NULL ) {
    LISTITEM_EXPR_NODE(varlist)->subnodetype = e_boxed;
    varlist = LISTNEXT_EXPR_NODE(varlist); 
  }
  DBUG_RETURN(res);
} //create_expr_param_list



Expr_node_ptr
create_expr_arg_list(Expr_node_ptr arg, Expr_node_ptr next)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_arg_list");
  DBUG_ASSERT(arg != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_arg_list;
  res->strrep             = NULL;
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = -1;
  COLNO_NODE(res)         = -(__LINE__);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARG_EXPR_NODE(res)  = arg;
  NEXT_EXPR_NODE(res) = next;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arg);
    tree_parts = myGlistRemove(tree_parts, next);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_arg_list


Expr_node_ptr
create_expr_listkeeper(struct _Mlval_struct value, Expr_node_ptr list)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_listkeeper");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_listkeeper;
  res->strrep             = mtstrdup("[ ]");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = value.lineno;
  COLNO_NODE(res)         = value.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LIST_EXPR_NODE(res)  = list;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, list);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_listkeeper

Expr_node_ptr
create_expr_list(Expr_node_ptr elm, Expr_node_ptr next)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_list");
  DBUG_ASSERT(elm != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_list;
  res->strrep             = NULL;
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = LINENO_NODE(elm);
  COLNO_NODE(res)         = COLNO_NODE( elm);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LISTITEM_EXPR_NODE(res)  = elm;
  LISTNEXT_EXPR_NODE(res) = next;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, elm);
    tree_parts = myGlistRemove(tree_parts, next);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_list

Expr_node_ptr
create_expr_product(Expr_node_ptr list)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_list");
  DBUG_ASSERT(list != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_product;
//  printf(sprintf("asdasd",1));
  res->strrep             = mtstrdup("( , )");//NULL;//sprintf("%s, %s", prd->strrep, next->strrep);
  DBUG_EXECUTE("strrep", res->strrep = mtstrdup(nodetype2str(res->nodetype)););
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = LINENO_NODE(list);
  COLNO_NODE(res)         = COLNO_NODE(list);
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LISTITEM_EXPR_NODE(res)  = list;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, list);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_product


Expr_node_ptr
create_expr_constant(struct _Mlval_struct value, MLType valuetype_found)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_constant");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_constant;

  //all about the value has to be set befor using mlConstant2str()
  res->value              = value.val; //it is a union!!!
  res->valuetype_expected = valuetype_found; //for safety
  res->valuetype          = valuetype_found;
  res->strrep             = mlConstant2String(res);//value has to set allready
  /* using this trick a mlstring will be freed too */
  if (valuetype_found == 5/*mlstring*/) {
    res->dummy_strval     = value.val.mlstring;
  } else {
    res->dummy_strval     = NULL;
  }
  res->misc               = NULL;
  LINENO_NODE(res)        = value.lineno;
  COLNO_NODE(res)         = value.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_constant


Expr_node_ptr
create_expr_nil_const(struct _Mlval_struct value, MLType valuetype_found)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_nil_const");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_nil_const;
  res->strrep             = mtstrdup("[]");
  res->value              = value.val; //it is a union!!!
  res->dummy_strval       = NULL;
  res->valuetype_expected = valuetype_found; //for safety
  res->valuetype          = valuetype_found;
  res->misc               = NULL;
  LINENO_NODE(res)        = value.lineno;
  COLNO_NODE(res)         = value.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_nil_const


Expr_node_ptr
create_expr_variable(struct _Strval_struct strval)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_variable");
  DBUG_ASSERT(strval.str != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_variable;
  res->strrep             = mtstrdup(strval.str);
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = strval.lineno;
  COLNO_NODE(res)         = strval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_variable


Expr_node_ptr
create_expr_specialtype_param(gchar* var_id, MLType valuetype_expected)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_specialtype_param");
  DBUG_ASSERT(var_id != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_variable;
  res->strrep             = mtstrdup(var_id);
  res->dummy_strval       = NULL;
  res->valuetype_expected = valuetype_expected; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = 0;
  COLNO_NODE(res)         = 0;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_specialtype_param

Expr_node_ptr
create_expr_producttype_param(gchar* var_id, MLType valuetype_expected)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_specialtype_param");
  DBUG_ASSERT(var_id != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_variable;
  res->strrep             = mtstrdup(var_id);
  res->dummy_strval       = NULL;
  res->valuetype_expected = valuetype_expected; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = 0;
  COLNO_NODE(res)         = 0;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  printf("CREATING PRODUCTTYPEPARAM\n");
  if (keep_tree_parts_in_mind) {
    printf("length %i\n", g_list_length(tree_parts));
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_producttype_param


Expr_node_ptr
create_expr_unary_not(struct _Noval_struct noval, Expr_node_ptr operand)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_unary_not");
  DBUG_ASSERT(operand != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype              = e_unary_not;
  res->strrep                = mtstrdup("not");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype             = 0; //0 means not set yet
  res->misc                  = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  OPERAND_EXPR_NODE(res) = operand;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, operand);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_unary_not


Expr_node_ptr
create_expr_unary_minus(struct _Noval_struct noval, Expr_node_ptr operand)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_unary_minus");
  DBUG_ASSERT(operand != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_unary_minus;
  res->strrep             = mtstrdup("~");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  OPERAND_EXPR_NODE(res) = operand;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, operand);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_unary_minus


Expr_node_ptr
create_expr_infixop(struct _Noval_struct noval, Nodetype type, 
                    Expr_node_ptr left_operand, 
                    Expr_node_ptr right_operand)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_infixop");
  DBUG_ASSERT(left_operand  != NULL);
  DBUG_ASSERT(right_operand != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype                = type;
  res->strrep                  = infixop2strrep(type);
  res->dummy_strval       = NULL;
  res->valuetype_expected      = 0; //0 means not set yet
  res->valuetype               = 0; //0 means not set yet
  res->misc                    = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LEFT_OPERAND_EXPR_NODE(res)  = left_operand;
  RIGHT_OPERAND_EXPR_NODE(res) = right_operand;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, left_operand);
    tree_parts = myGlistRemove(tree_parts, right_operand);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_infixop


Expr_node_ptr
create_expr_basicfun(struct _Noval_struct noval, 
                     Nodetype type, Expr_node_ptr arg)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_basicfun");
  DBUG_ASSERT(arg != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = type;
  res->strrep             = basicfun2strrep(type);
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARG_EXPR_NODE(res)  = arg;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arg);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_basicfun


Expr_node_ptr
create_expr_basicfun2(struct _Noval_struct noval, 
		      Nodetype type, Expr_node_ptr arg1, Expr_node_ptr arg2)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_basicfun2");
  DBUG_ASSERT(arg1 != NULL);
  DBUG_ASSERT(arg2 != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = type;
  res->strrep             = basicfun2strrep(type);
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LEFT_OPERAND_EXPR_NODE(res)  = arg1;
  RIGHT_OPERAND_EXPR_NODE(res) = arg2;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arg1);
    tree_parts = myGlistRemove(tree_parts, arg2);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_basicfun2


Expr_node_ptr
create_expr_substring(struct _Noval_struct noval, Expr_node_ptr str, 
                      Expr_node_ptr startpos, Expr_node_ptr length)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_substring");
  DBUG_ASSERT(str      != NULL);
  DBUG_ASSERT(startpos != NULL);
  DBUG_ASSERT(length   != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_substring;
  res->strrep             = mtstrdup("substring");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  res->child[0] = str;
  res->child[1] = startpos;
  res->child[2] = length;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, str);
    tree_parts = myGlistRemove(tree_parts, startpos);
    tree_parts = myGlistRemove(tree_parts, length);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_substring


Expr_node_ptr
create_expr_project(Expr_node_ptr pos,     
                       Expr_node_ptr expr)
{
  Expr_node_ptr res;
  int i;
  char *buffer = mtmalloc(50);

  DBUG_ENTER("create_expr_project");
  DBUG_ASSERT(pos != NULL);
  DBUG_ASSERT(expr != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype            = e_project;
  sprintf(buffer,"%d",(pos->value.mlint)+1);
  res->strrep              = mtstrconcat(2,mtstrdup("#"),mtstrdup(buffer));
  res->dummy_strval       = NULL;
  res->valuetype_expected  = 0; //0 means not set yet
  res->valuetype           = 0; //0 means not set yet
  res->misc                = NULL;
 // LINENO_NODE(res)         = noval.lineno;
 // COLNO_NODE(res)          = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  LEFT_OPERAND_EXPR_NODE(res)      = pos;
  RIGHT_OPERAND_EXPR_NODE(res)      = expr;
  pos->subnodetype = e_hidden;


  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, pos);
    tree_parts = myGlistRemove(tree_parts, expr);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_ifthenelse

Expr_node_ptr
    create_expr_ifthenelse(struct _Noval_struct noval, Expr_node_ptr cond,     
                           Expr_node_ptr then_expr, Expr_node_ptr else_expr)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_ifthenelse");
  DBUG_ASSERT(cond      != NULL);
  DBUG_ASSERT(then_expr != NULL);
  DBUG_ASSERT(else_expr != NULL);

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype            = e_ifthenelse;
  res->strrep              = mtstrdup("if");
  res->dummy_strval       = NULL;
  res->valuetype_expected  = 0; //0 means not set yet
  res->valuetype           = 0; //0 means not set yet
  res->misc                = NULL;
  LINENO_NODE(res)         = noval.lineno;
  COLNO_NODE(res)          = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  COND_EXPR_NODE(res)      = cond;
  THEN_EXPR_EXPR_NODE(res) = then_expr;
  ELSE_EXPR_EXPR_NODE(res) = else_expr;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, cond);
    tree_parts = myGlistRemove(tree_parts, then_expr);
    tree_parts = myGlistRemove(tree_parts, else_expr);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_ifthenelse



Expr_node_ptr
create_expr_constructor(struct _Strval_struct  strval, Expr_node_ptr arglist)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_constructor");
  //arglist may be NULL

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_constructor;
  res->strrep             = mtstrdup(strval.str);
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = strval.lineno;
  COLNO_NODE(res)         = strval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARGLIST_EXPR_NODE(res) = arglist;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arglist);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_constructor


Expr_node_ptr
create_expr_userfun(struct _Strval_struct  strval, Expr_node_ptr arglist)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_userfun");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_userfun;
  res->strrep             = mtstrdup(strval.str);
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = strval.lineno;
  COLNO_NODE(res)         = strval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }
  ARGLIST_EXPR_NODE(res) = arglist;

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, arglist);
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_userfun


Expr_node_ptr
create_expr_wildcard(struct _Noval_struct noval)
{
  Expr_node_ptr res;
  int i;

  DBUG_ENTER("create_expr_wildcard");

  res = mtmalloc(sizeof(Expr_node_struct));
  res->nodetype           = e_wildcard;
  res->strrep             = mtstrdup("_");
  res->dummy_strval       = NULL;
  res->valuetype_expected = 0; //0 means not set yet
  res->valuetype          = 0; //0 means not set yet
  res->misc               = NULL;
  LINENO_NODE(res)        = noval.lineno;
  COLNO_NODE(res)         = noval.colno;
  res->expansiondepth     = 0;
  for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
    res->child[i] = NULL;
  }

  if (keep_tree_parts_in_mind) {
    tree_parts = g_list_prepend(tree_parts, res);
  }

  DBUG_RETURN(res);
} //create_expr_wildcard


/** Frees the ressources used by an Expr_node but not the contents of 
  * node->misc. So node->misc has to be NULL if free_expr_node() is called.
  * Another way to free node->misc is to set the global function pointer 
  * fpFree_expr_node_misc befor free_expr_node() is called.
  * The children are freed too. Otherwise they have to be NULL.
  * @return NULL so you can free a pointer an set it to NULL in one line
  *         ( node = free_expr_node(node); ).
  */
Expr_node_ptr
free_expr_node(Expr_node_ptr node)
{
  int i;

  DBUG_ENTER("free_expr_node");

  if (node != NULL) {
    //printf_expr_node(node);
    //free its children
    for (i=0; i < EXPR_NODE_MAX_CHILDREN; i++) {
      node->child[i] = free_expr_node(node->child[i]);
    } //for i

    //free itself
    MTFREENULL(node->strrep);
    MTFREENULL(node->dummy_strval);
    if ((fpFree_expr_node_misc != NULL) && (node->misc != NULL)) {
      (*fpFree_expr_node_misc)(node->misc);
      node->misc = NULL;
    }
    DBUG_ASSERT(node->misc == NULL);
    MTFREENULL(node);  
  } else {
    DBUG_PRINT("free", ("expr_node is allready NULL"));
  }

  DBUG_RETURN((Expr_node_ptr)NULL);
} //free_expr_node


/** Create_fun_node allocate memory and set the node->nodetype, the params
  * and the body, and all the rest to defaults (0 or NULL).
  * @return a pointer to the fresh allocated Fun_node_struct.
  */
Fun_node_ptr
create_fun_node(gchar* fun_id, Nodetype nodetype, Expr_node_ptr params,
                MLType resulttype, Expr_node_ptr body, 
                GList* bindings)
{
  Fun_node_ptr res;

  DBUG_ENTER("create_fun_node");
  DBUG_ASSERT( (nodetype == e_invalid) || (nodetype == e_userfun) 
               || (nodetype == e_constructor) );
  DBUG_ASSERT(fun_id != NULL);
  //params == NULL in case of a parameterless constructor
  //body   == NULL in case of a constructor

  res = mtmalloc(sizeof(Fun_node_struct));
  res->fun_id       = mtstrdup(fun_id);
  res->nodetype     = nodetype;
  res->resulttype   = resulttype;
  res->paramslist   = g_list_append(res->paramslist,   params);
  res->bindingslist = g_list_append(res->bindingslist, bindings);
  res->bodylist     = g_list_append(res->bodylist,     body);

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, params);
    tree_parts = myGlistRemove(tree_parts, body);
  }

  DBUG_RETURN(res);
} //create_fun_node


/* Appand another clause to a fun_node.
 */
void
append_fun_node(Fun_node_ptr pFunnode, 
		Expr_node_ptr params, GList* bindings, Expr_node_ptr body)
{
  DBUG_ENTER("append_fun_node");

  /* It is important to append the values instead of prepending them, 
   * because of the patternmatching matches the first clause found */
  pFunnode->paramslist   = g_list_append(pFunnode->paramslist,   params);
  pFunnode->bindingslist = g_list_append(pFunnode->bindingslist, bindings);
  pFunnode->bodylist     = g_list_append(pFunnode->bodylist,     body);

  if (keep_tree_parts_in_mind) {
    tree_parts = myGlistRemove(tree_parts, params);
    tree_parts = myGlistRemove(tree_parts, body);
  }

  DBUG_VOID_RETURN;
} //append_fun_node


void
printf_fun_node(Fun_node_ptr node)
{
  GList *gl1, *gl2;

  DBUG_ENTER("printf_fun_node");

  printf("Fun_Node_ptr: %s (%p)\n", node->fun_id, node);
  gl1 = node->paramslist;
  gl2 = node->bodylist;
  while ((gl1 != NULL) && (gl2 != NULL)) {
    printf("- params: %p\tbody : %p\n", gl1->data, gl2->data);
    gl1 = g_list_next(gl1);
    gl2 = g_list_next(gl2);
  }

  DBUG_VOID_RETURN;
} //printf_fun_node


/** Frees the ressources used by a Fun_node even the contents of 
  * node->params and node->body. Otherwise they have to be  NULL.
  * @return NULL so you can free a pointer an set it to NULL in one line
  *         ( node = free_fun_node(node); ).
  */
Fun_node_ptr
free_fun_node(Fun_node_ptr node)
{
  GList *gl1, *gl2, *gl3;

  DBUG_ENTER("free_fun_node");

  if (node != NULL) {
    //free its children
    gl1 = node->paramslist;
    gl2 = node->bindingslist;
    gl3 = node->bodylist;
    while (gl1 != NULL) {
      gl1->data = free_expr_node(gl1->data);
      gl1 = g_list_next(gl1);
    }
    g_list_free(gl1);
    while (gl2 != NULL) {
      gl2->data = free_symboltable(gl2->data);
      gl2 = g_list_next(gl2);
    }
    g_list_free(gl2);
    while (gl3 != NULL) {
      gl3->data = free_expr_node(gl3->data);
      gl3 = g_list_next(gl3);
    }
    g_list_free(gl3);

    //free itself
    MTFREENULL(node->fun_id);
    MTFREENULL(node);  
  }
  
  DBUG_RETURN((Fun_node_ptr)NULL);
} //free_fun_node


GList*
free_syntaxtree(GList* st)
{
  GList* tmp;

  DBUG_ENTER("free_syntaxtree");

  for (tmp = st;  tmp != NULL;  tmp = g_list_next(tmp)) {
    tmp->data = free_fun_node(tmp->data);
  } //for tmp

  g_list_free(st);

  DBUG_RETURN((GList*)NULL);
} //free_syntaxtree


void
printf_treeparts(GList* tree_parts)
{
  GList* tmp;

  DBUG_ENTER("printf_treeparts");
  DBUG_PRINT( "try", ("printing tree_parts (nr_elements: %i)", 
                      g_list_length(tree_parts)) );

  for (tmp = tree_parts;  tmp != NULL;  tmp = g_list_next(tmp)) {
    printf_expr_node(tmp->data);
  } //for tmp

  DBUG_PRINT( "try", ("end tree_parts\n") );
  DBUG_VOID_RETURN;
} //printf_treeparts


GList*
free_treeparts(GList* tree_parts)
{
  GList* tmp;

  DBUG_ENTER("free_treeparts");
  DBUG_PRINT( "try", ("freeing tree_parts (nr_elements: %i)\n", 
                      g_list_length(tree_parts)) );

  for (tmp = tree_parts;  tmp != NULL;  tmp = g_list_next(tmp)) {
    tmp->data = free_expr_node(tmp->data);
  } //for tmp

  g_list_free(tree_parts);

  DBUG_RETURN((GList*)NULL);
} //free_treeparts


void nav_expr_node(Expr_node_ptr node)
{
  gint doExit = 0;
  gchar inputkey[2] = {0,0};

  DBUG_ENTER("nav_expr_node");

  while (!doExit) {
    if (node == NULL) {
      fprintf(stdout, "Expr_node_ptr is NULL!");
    } else {
      printf_expr_node(node);
    }
    fprintf(stdout, "\n(0) back   (1) child 1   (2) child 2   "
                    "(3) child 3 :");
    scanf("%1s", inputkey);
    switch (inputkey[0]) {
    case '0': doExit = 1; 
              break;
    case '1': nav_expr_node(node->child[0]);
              break;
    case '2': nav_expr_node(node->child[1]);
              break;
    case '3': nav_expr_node(node->child[2]);
              break;
    default : ;
    } //switch
  } //loop

  DBUG_VOID_RETURN;
} //nav_expr_node


static void nav_bodylist(GList* bl)
{
  gint doExit = 0;
  gchar inputkey[2] = {0,0};
  Expr_node_ptr pExpr;
  GList* newBl;

  DBUG_ENTER("nav_bodylist");

  while (!doExit) {
    if (bl == NULL) {
      fprintf(stdout, "GList (bodylist) is NULL!");
    } else {
      pExpr = bl->data;
      fprintf(stdout, "GList: nr_elements: %i\n",  g_list_length(bl));
      fprintf(stdout, "(GList->data) body: %p\n", pExpr);
      //printf_nodetype(pExpr);
    }
    fprintf(stdout, "\n(0) back   (-) previous clause   (+) next clause   "
                    "(b) body :");
    scanf("%1s", inputkey);
    newBl = bl;
    switch (inputkey[0]) {
    case '0': doExit = 1; 
              break;
    case '-': newBl = g_list_previous(bl);
              break;
    case '+': newBl = g_list_next(bl);
              break;
    case 'b': nav_expr_node(bl->data);
              break;
    default : ;
    } //switch
    if (newBl != NULL) { bl = newBl; }
  } //loop

  DBUG_VOID_RETURN;
} //nav_bodylist


static void nav_paramslist(GList* pl)
{
  gint doExit = 0;
  gchar inputkey[2] = {0,0};
  Expr_node_ptr pExpr;
  GList* newPl;

  DBUG_ENTER("nav_paramslist");

  while (!doExit) {
    if (pl == NULL) {
      fprintf(stdout, "GList (paramslist) is NULL!");
    } else {
      pExpr = pl->data;
      fprintf(stdout, "GList: nr_elements: %i\n",  g_list_length(pl));
      fprintf(stdout, "(GList->data) params: %p\n", pExpr);
      //printf_nodetype(pExpr);
    }
    fprintf(stdout, "\n(0) back   (-) previous clause   (+) next clause   "
                    "(p) params :");
    scanf("%1s", inputkey);
    newPl = pl;
    switch (inputkey[0]) {
    case '0': doExit = 1; 
              break;
    case '-': newPl = g_list_previous(pl);
              break;
    case '+': newPl = g_list_next(pl);
              break;
    case 'p': nav_expr_node(pl->data);
              break;
    default : ;
    } //switch
    if (newPl != NULL) { pl = newPl; }
  } //loop

  DBUG_VOID_RETURN;
} //nav_paramslist


void nav_syntaxtree(GList* st)
{
  gint doExit = 0;
  gchar inputkey[2] = {0,0};
  GList* newSt;

  DBUG_ENTER("nav_syntaxtree");

  while (!doExit) {
    if (st == NULL) {
      fprintf(stdout, "syntax_tree is NULL!");
    } else {
      printf_fun_node(st->data);
    }
    fprintf(stdout, "\n(x) exit   (-) previous fun   (+) next fun   "
                    "(p) paramslist   (b) bodylist :");
    scanf("%1s", inputkey);
    newSt = st;
    switch (inputkey[0]) {
    case 'x': case '0':
              doExit = 1; 
              break;
    case '-': newSt = g_list_previous(st);
              break;
    case '+': newSt = g_list_next(st);
              break;
    case 'p': if (st != NULL) { 
                nav_paramslist( ((Fun_node_ptr)st->data)->paramslist );
              }
              break;
    case 'b': if (st != NULL) { 
                nav_bodylist( ((Fun_node_ptr)st->data)->bodylist );
              }
              break;
    default : ;
    } //switch
    if (newSt != NULL) { st = newSt; }
  } //loop

  DBUG_VOID_RETURN;
} //nav_syntaxtree


gchar* 
funNode2Signature(Fun_node_ptr pFun)
{
  gchar* res;
  gchar tmpstr[256];
  gboolean addStar;
  Expr_node_ptr params;
  Expr_node_ptr pParam;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("funNode2Signature");
  DBUG_ASSERT(pFun != NULL);

  //printf_fun_node(pFun);
  tmpstr[0] = '\0';
  if        (pFun->nodetype == e_constructor) {
    sprintf(tmpstr, "  con %s : ", pFun->fun_id);
    params = pFun->paramslist->data;
    addStar = FALSE;
    while (params != NULL) {
      if (addStar) {
	strcat(tmpstr, " * ");
      } else {
	//strcat(tmpstr, " of ");
      }
      pParam = PARAM_EXPR_NODE(params);
      pSte = seekSymbolTypeindex(symbol_table, pParam->valuetype);
      DBUG_ASSERT(pSte != NULL);
      strcat(tmpstr, pSte->id);

      addStar = TRUE;
      params = NEXT_EXPR_NODE(params);
    } //while
    if (pFun->paramslist->data != NULL) {
      strcat(tmpstr, " -> ");
    }
    pSte = seekSymbolTypeindex(symbol_table, pFun->resulttype);
    DBUG_ASSERT(pSte != NULL);
    strcat(tmpstr, pSte->id);


  } else if (pFun->nodetype == e_userfun) {
    sprintf(tmpstr, "fun %.250s : ", pFun->fun_id);
    params = pFun->paramslist->data;
    if (params != NULL) {
      addStar = FALSE;
      while (params != NULL) {
	if (addStar) {
	  strcat(tmpstr, " * ");
	}
	pParam = PARAM_EXPR_NODE(params);
	pSte = seekSymbolTypeindex(symbol_table, pParam->valuetype);
	DBUG_ASSERT(pSte != NULL);
	strcat(tmpstr, pSte->id);
	
	addStar = TRUE;
	params = NEXT_EXPR_NODE(params);
      }
      strcat(tmpstr, " -> ");
    } //if params != NULL
    pSte = seekSymbolTypeindex(symbol_table, pFun->resulttype);
    DBUG_ASSERT(pSte != NULL);
    strcat(tmpstr,  pSte->id);
  } else {
    DBUG_ASSERT(0); //unkown or wrong nodetype
  }

  res = mtstrdup(tmpstr);

  DBUG_RETURN(res);
} //funNode2Signature


gchar*
funNode2DatatypeStr(Fun_node_ptr pFun)
{
  gchar* res = NULL;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER(funNode2DatatypeStr);
  DBUG_ASSERT(pFun != NULL);

  if (pFun->nodetype == e_constructor) {
    pSte = seekSymbolTypeindex(symbol_table, pFun->resulttype);
    DBUG_ASSERT(pSte != NULL);
    res =mtstrconcat(2, "datatype ", pSte->id);
  } //if

  DBUG_RETURN(res);
} //funNode2DatatypeStr


static gchar* 
exprNode2StringRec(Expr_node_ptr pExpr, TIsFirstElmInList isFirst)
{
  gchar* res = NULL;
  gchar *str_c1, *str_c2, *str_c3;

  DBUG_ENTER("exprNode2StringRec");
  DBUG_ASSERT(pExpr != NULL);


  switch (pExpr->nodetype) {
  case e_errornode:
    DBUG_PRINT("2String", ("e_errornode"));
    res = mtstrdup(" ERROR ");
    break;

  case e_constant:
    DBUG_PRINT("2String", ("e_constant"));
    res = mlConstant2String(pExpr);
    break;


  case e_nil_const:
    DBUG_PRINT("2String", ("e_nil_const"));
    res = mtstrdup("nil");
    break;

  case e_variable:
    DBUG_PRINT("2String", ("e_variable"));
    res = mtstrdup(pExpr->strrep);
    break;

  case e_param_list:
  case e_arg_list:
    DBUG_PRINT("2String", ("e_paramlist+"));
    str_c1 = exprNode2StringRec(PARAM_EXPR_NODE(pExpr), e_yes);
    if (NEXT_EXPR_NODE(pExpr) != NULL) {
      str_c2 = exprNode2StringRec(NEXT_EXPR_NODE(pExpr), e_no);
    } else {
      str_c2 = NULL;
    }

    if        (isFirst == e_yes) {
      res = mtstrconcat(2, str_c1, str_c2);
    } else if (isFirst == e_no) {
      res = mtstrconcat(3, ", ", str_c1, str_c2);
    } else {
      DBUG_ASSERT(0);
    }
    mtmultifree(2, str_c1, str_c2);
    break;

 
  case e_listkeeper:
    DBUG_PRINT("2String", ("e_listkeeper"));
    if (LIST_EXPR_NODE(pExpr) == NULL) {
      res = mtstrdup(pExpr->strrep);
    } else {
      res = exprNode2StringRec(LIST_EXPR_NODE(pExpr), e_yes);
    }
    break;


  case e_list:
    DBUG_PRINT("2String", ("e_list"));
    if (NEXT_EXPR_NODE(pExpr) != NULL) {
      str_c1 = exprNode2StringRec(LISTITEM_EXPR_NODE(pExpr), e_yes);
      str_c2 = exprNode2StringRec(LISTNEXT_EXPR_NODE(pExpr), e_no);
    } else {
      str_c1 = exprNode2StringRec(LISTITEM_EXPR_NODE(pExpr), e_yes);
      str_c2 = mtstrdup("]");
    }

    if        (isFirst == e_yes) {
      res = mtstrconcat(3, "[", str_c1, str_c2);
    } else if (isFirst == e_no) {
      if (str_c1 == NULL) {
	res = mtstrdup(str_c2);
      } else {
	res = mtstrconcat(3, ", ", str_c1, str_c2);
      }
    } else {
      DBUG_ASSERT(0);
    }
    mtmultifree(2, str_c1, str_c2);
    break;

    case e_let:
      str_c1 = exprNode2StringRec(LET_EXPR_EXPR_NODE(pExpr), e_yes);
      str_c2 = exprNode2StringRec(LET_VAL_EXPR_NODE(pExpr), e_no);
      str_c3 = exprNode2StringRec(LET_VARS_EXPR_NODE(pExpr), e_no);

      res = mtstrconcat(7, mtstrdup("let ("),str_c3,mtstrdup(") = "), str_c2, mtstrdup(" in "), str_c1, mtstrdup(" end"));
      break;
      
    case e_letvars_list:
      DBUG_PRINT("2String", ("e_letvars_list"));
      
      res = NULL;
{
      Expr_node_ptr vars = pExpr;
      while ( vars != NULL ) {
        Symboltable_entry_ptr pSteE = seekSymbolTypeindex(symbol_table, ARG_EXPR_NODE(vars)->valuetype_expected);
        if ( res == NULL ) {
          res = mtstrconcat(3,mtstrdup(ARG_EXPR_NODE(vars)->strrep),mtstrdup(":"),mtstrdup(pSteE->id));
        } else {
          res = mtstrconcat(5,res,mtstrdup(","),mtstrdup(ARG_EXPR_NODE(vars)->strrep),mtstrdup(":"),mtstrdup(pSteE->id));
        }
        vars = NEXT_EXPR_NODE(vars);
      }
}      
      break;

  case e_product:
    DBUG_PRINT("2String", ("e_product"));
{    
    Expr_node_ptr list = LISTITEM_EXPR_NODE(pExpr);
    res = NULL;
    while ( list != NULL ) {
      if ( res == NULL ) {
        res = mtstrconcat(2,mtstrdup("("),exprNode2StringRec(LISTITEM_EXPR_NODE(list), e_yes));
      } else {
        res = mtstrconcat(3,res,mtstrdup(","),exprNode2StringRec(LISTITEM_EXPR_NODE(list), e_yes));

      }
      list = LISTNEXT_EXPR_NODE(list);
    }
    res = mtstrconcat(2,res,mtstrdup(")"));
}
    break;



  /* case e_variable: break;
   * there are no global variables allowd */


  case e_unary_not:
    str_c1 = exprNode2StringRec(OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(2, "not ", str_c1);
    MTFREENULL(str_c1);
    break;


  case e_unary_minus:
    str_c1 = exprNode2StringRec(OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(2, "~", str_c1);
    MTFREENULL(str_c1);
    break;


  case e_andalso:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " andalso ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_orelse:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " orelse ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_equal:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " = ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_unequal:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " <> ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_lt:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " < ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_le:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " <= ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;

  case e_project:
      str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
      str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
      res = mtstrconcat(5, "project(", str_c1, ", ", str_c2, ")");
      mtmultifree(2, str_c1, str_c2);
      break;


  case e_gt:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " > ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_ge:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " >= ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_plus:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " + ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_minus:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " - ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_mult:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " * ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_mod:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " mod ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_int_div:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " div ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_real_div:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " / ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_string_concat:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " ^ ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_prepend:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " :: ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_list_concat:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "(", str_c1, " @ ", str_c2, ")");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_abs:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "abs( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_min:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr),  e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "min( ", str_c1, ", ", str_c2, " )");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_max:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(RIGHT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(5, "max( ", str_c1, ", ", str_c2, " )");
    mtmultifree(2, str_c1, str_c2);
    break;


  case e_sqrt:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "sqrt( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_sin:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "sin( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_cos:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "cos( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_exp:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "exp( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_ln:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "ln( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_real:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "real( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_truncate:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "truncate( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_size:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "size( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_ord:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "ord( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_chr:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "chr( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_substring:
    str_c1 = exprNode2StringRec(pExpr->child[0], e_yes);
    str_c2 = exprNode2StringRec(pExpr->child[1], e_yes);
    str_c3 = exprNode2StringRec(pExpr->child[2], e_yes);
    res = mtstrconcat(7, "substring( ", str_c1, ", ", str_c2, ", ", 
		         str_c3, " )");
    mtmultifree(3, str_c1, str_c2, str_c3);
    break;


  case e_null:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "null( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_hd:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "hd( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_tl:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "tl( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_length:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "length( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_rev:
    str_c1 = exprNode2StringRec(LEFT_OPERAND_EXPR_NODE(pExpr), e_yes);
    res = mtstrconcat(3, "rev( ", str_c1, " )");
    MTFREENULL(str_c1);
    break;


  case e_ifthenelse:
    str_c1 = exprNode2StringRec(COND_EXPR_NODE(pExpr), e_yes);
    str_c2 = exprNode2StringRec(THEN_EXPR_EXPR_NODE(pExpr), e_yes);
    str_c3 = exprNode2StringRec(ELSE_EXPR_EXPR_NODE(pExpr), e_yes);
    res=mtstrconcat(7,"(if ", str_c1, " then ", str_c2, " else ", str_c3, ")");
    mtmultifree(3, str_c1, str_c2, str_c3);
    break;

  case e_constructor:
  case e_userfun:
    if (PARAMLIST_EXPR_NODE(pExpr) == NULL) {
      res = mtstrconcat(2, pExpr->strrep,
                           (pExpr->nodetype == e_userfun ? "()":""));
    } else {
      str_c1 = exprNode2StringRec(PARAMLIST_EXPR_NODE(pExpr), e_yes);
      res = mtstrconcat(4, pExpr->strrep, "( ", str_c1, " )");
      MTFREENULL(str_c1);
    }
    break;



  /* case e_wildcard: break;
   * there are no wildcards allowd */


  default:
    printf_expr_node(pExpr);
    DBUG_ASSERT(0);
  } //switch

  DBUG_PRINT("2String", ("res = '%s' (%p)", res, res));
  DBUG_RETURN(res);
} //exprNode2StringRec


gchar* 
exprNode2String(Expr_node_ptr pExpr)
{ gchar* res;

  DBUG_ENTER("exprNode2String");
  DBUG_ASSERT(pExpr != NULL);

  res = exprNode2StringRec(pExpr, e_yes);

  DBUG_RETURN(res);
} //exprNode2String


/* @return -1 if 'child' is not a child of 'parent'
 *         or the index otherwise
 */
gint 
getIndexOfChild(Expr_node_ptr parent, Expr_node_ptr child)
{
  gint i, res = -1;

  DBUG_ENTER("getIndexOfChild");
  DBUG_ASSERT(parent != NULL);

  i = 0;
  while (i < EXPR_NODE_MAX_CHILDREN) {
    if (parent->child[i] == child) {
      res = i;
      i = EXPR_NODE_MAX_CHILDREN+2;
    } else {
      i++;
    }
  } //while

  DBUG_RETURN(res);
} //getIndexOfChild


/* Frees the whole subtree of old and overwrites the old in parent_of_old
 * with the new.
 * If parent_of_old is not the parent of old nothing happens.
 */
void 
replaceSubtree(Expr_node_ptr parent_of_old, Expr_node_ptr old, 
	       Expr_node_ptr new)
{
  gint index;

  DBUG_ENTER("replaceSubtree");
  DBUG_ASSERT(parent_of_old != NULL);

  index = getIndexOfChild(parent_of_old, old);
  if (index >= 0) {
    parent_of_old->child[index] = new;
    free_expr_node(old);
    old = NULL;
  } else {
    DBUG_PRINT("", ("nothing replaced!"));
  }

  DBUG_VOID_RETURN;
} //replaceSubtree


/* Duplicates a single expr-node expect the pointers to its children and
 * the ->misc. They are set to NULL.
 * @return the new expr-node
 */
Expr_node_ptr 
duplicateSingleExprnode(Expr_node_ptr src)
{
  Expr_node_ptr res;
  gint i;

  DBUG_ENTER("duplicateSingleExprnode");

  if (src == NULL) {
    res = NULL;
  } else {
    res = mtmalloc(sizeof(Expr_node_struct));
    res->nodetype           = src->nodetype;
    res->strrep             = mtstrdup(src->strrep);
    res->subnodetype        = src->subnodetype;
    if (src->dummy_strval != NULL) {
      //it is a mlstring
      res->dummy_strval       = mtstrdup(src->dummy_strval);
      res->value.mlstring     = res->dummy_strval;
    } else {
      res->value              = src->value;
    }
    res->valuetype_expected = src->valuetype_expected;
    res->valuetype          = src->valuetype;
    res->misc               = NULL;
    LINENO_NODE(res)        = LINENO_NODE(src);
    COLNO_NODE(res)         = COLNO_NODE(src);
    res->expansiondepth     = src->expansiondepth;
    for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
      res->child[i] = NULL;
    } //for i
  } //if

  DBUG_RETURN(res);
} //duplicateSingleExprnode


Expr_node_ptr 
duplicateSubtree(Expr_node_ptr src)
{
  Expr_node_ptr res;
  gint i;

  DBUG_ENTER("duplicateSubtree");

  if (src == NULL) {
    res = NULL;
  } else {
    res = duplicateSingleExprnode(src);
    for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
      res->child[i] = duplicateSubtree(src->child[i]);
    } //for i
  } //if

  DBUG_RETURN(res);
} //duplicateSubtree


Expr_node_ptr 
duplicateSubtreeButSetNewExpansiondepth(Expr_node_ptr src, 
					gint new_expansiondepth)
{
  Expr_node_ptr res;
  gint i;

  DBUG_ENTER("duplicateSubtreeButSetNewExpansiondepth");

  if (src == NULL) {
    res = NULL;
  } else {
    res = duplicateSingleExprnode(src);
    res->expansiondepth = new_expansiondepth;
    for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
      res->child[i] = 
	duplicateSubtreeButSetNewExpansiondepth(src->child[i], 
						new_expansiondepth);
    } //for i
  } //if

  DBUG_RETURN(res);
} //duplicateSubtreeButSetNewExpansiondepth

