#include <stdio.h>
#include <gtk/gtk.h>
#include "../Misc/my_dbug.h"
//#include "../Misc/error_interface.h"
#include "../Syntaxtree/symboltable.h"
#include "../Syntaxtree/syntaxtree.h"
#include "tc_util.h"
#include "tc_body.h"

typedef gint (*Check_funptr) (Expr_node_ptr, gint);
typedef gint (*NodeAction_funptr) (GList*, Expr_node_ptr, Expr_node_ptr, gint);

/*****************************************************************************
 *
 * Some functions for the cases, which have to be handled the same way.
 * The pointer to a function called fpRecursivCall has to call recursivly
 * the calling function of case_e_XYZ, which may be checkParams or
 * checkBody.
 */

static gint
case_e_list(Check_funptr fpRecursivCall, Expr_node_ptr node)
{
  gint res = 0;
  MLType listitem_type;

  DBUG_ENTER("case_e_list");

  DBUG_ASSERT(node->valuetype_expected > 0);

  //is type really a MLList?
  listitem_type = getItemtypeOfList(node->valuetype_expected);
  //printf("listitem_type: %d   expected: %d\n", listitem_type, node->valuetype_expected);
  DBUG_ASSERT(listitem_type > -1); //-1 means type not found in symbol_table
  if (listitem_type == 0) {
    //type is not a list
    tc_error(LINENO_NODE(node), COLNO_NODE(node),
	     node->valuetype_expected, list_type_index);
    printf("111-%p\n", node);
    res = 1;
  } else {
    /* set valuetype to the expected */
    node->valuetype = node->valuetype_expected;

    /* only check the relavant children */
    res = (*fpRecursivCall)(LISTITEM_EXPR_NODE(node), listitem_type);
    if (res == 0) {
      res = (*fpRecursivCall)(LISTNEXT_EXPR_NODE(node), node->valuetype);
    }
  }

  DBUG_RETURN(res);
} //case_e_list


static gint
case_e_constant(Check_funptr fpRecursivCall, Expr_node_ptr node)
{
  gint res = 0;

  DBUG_ENTER("case_e_constant");

  /* check if expected type is equal to the type of the constant */
  if ((node->valuetype_expected != 0) 
      && (node->valuetype_expected != node->valuetype)) {
    tc_error(LINENO_NODE(node), COLNO_NODE(node), 
	     node->valuetype_expected, node->valuetype);
    res = 1;
  }

  DBUG_RETURN(res);
} //case_e_constant


/*
 * End of case_XYZ functions
 *
 ****************************************************************************/


/** Check the pattern in a parameter in a top-down analysis (remember that 
  * the root-node of this subtree is allready typed). A bottom-up
  * analysis is not possible, because the leefs of the parameter subtree
  * are constants (which type is known) and variables, which type is still
  * unknown!
  * @param node is the actual node which has to be checked.
  * @param parent_expects is the MLType which the parent of the actual node
  *        expects this node to have. 
  * @return 0, but 1 in case of an error.
  */
static gint
checkParams(Expr_node_ptr node, MLType parent_expects)
{
  gint actparam_count, res = 0;
  MLType listitem_type ;
  Expr_node_ptr pActParams, pDefinedParams;
  Fun_node_ptr pFNode;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("checkParams");

  if (node == NULL) {
    DBUG_RETURN(res);
  }
  
  //If the parent node expects a type>0, then overwrite the 
  //valuetype_expected of the son.
  if (parent_expects > 0) {
    node->valuetype_expected = parent_expects;
  }

  switch (node->nodetype) {
  case e_wildcard:
    DBUG_PRINT( "tc", ("e_wildcard - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    /* set valuetype to the expected one */
    node->valuetype = node->valuetype_expected;
    break;

  case e_constant:
    DBUG_PRINT( "tc", ("e_constant - %s (%d)", 
      seekSymbolTypeindex(symbol_table, node->valuetype_expected)->id, 
		          node->valuetype_expected) );
    res = case_e_constant(&checkParams, node);
    break;

  case e_nil_const:
    DBUG_PRINT( "tc", ("e_nil_const - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
		       node->valuetype_expected) );
    DBUG_ASSERT(0); //e_nil_const should not occur anymore
    if (node->valuetype_expected != 0) {
      node->valuetype = node->valuetype_expected;
    }
    res = 0;
    break;

  case e_variable:
    DBUG_PRINT( "tc", ("e_variable - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    /* set valuetype to the expected one */
    DBUG_ASSERT(node->valuetype_expected > 0);
    node->valuetype = node->valuetype_expected;
    /* update type_index of corresponding Symboltable_entry in case of a fun */
    if (act_binding != NULL) {
      //it is a fun
      pSte = seekSymbolId(act_binding, node->strrep);
      DBUG_ASSERT(pSte != NULL);
      if (pSte->type_index == 0) {
	pSte->type_index = node->valuetype;
      } else {
	DBUG_ASSERT(pSte->type_index == node->valuetype);
      }
    } // else it is a constructor
    break;

  case e_constructor:
    DBUG_PRINT( "tc", ("e_constructor - %s (%d)",seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    /* A constructor allready owns parameters of a known type. */

    pSte = seekSymbolId(symbol_table, node->strrep);
    DBUG_ASSERT(pSte != NULL); //Seek must never fail!!!

    /* check resulttype of the constructor */
    if ( (node->valuetype_expected != 0) &&
         (node->valuetype_expected != pSte->type_index) ) {
      tc_error(LINENO_NODE(node), COLNO_NODE(node),
               node->valuetype_expected, pSte->type_index);
      res = 1;        
    } else {
      node->valuetype = pSte->type_index;
    } //if
    DBUG_PRINT( "tc",("evaluated type is %s (%d)", 
                      (seekSymbolTypeindex(symbol_table, node->valuetype))->id,
                      node->valuetype));

    
    /* check the parameter-list of the constructor
     * and if it is realy a e_param_list */
    pActParams = PARAMLIST_EXPR_NODE(node);
    if (pActParams != NULL) {
      if (pActParams->nodetype != e_param_list) {
        tc_error(LINENO_NODE(node), COLNO_NODE(node),
                 e_param_list, pActParams->nodetype);
        res = 1;
      } else {
        pFNode = pSte->definition_point;
        DBUG_ASSERT(pFNode != NULL);
        DBUG_ASSERT(pFNode->paramslist != NULL);
        pDefinedParams = pFNode->paramslist->data; //a constructor has only 
	                                           //one clause
        actparam_count = 0;
        while ( (pDefinedParams != NULL) && (pActParams != NULL)
                && (res == 0) ) {
          actparam_count += 1;
          res = checkParams(PARAM_EXPR_NODE(pActParams), 
                            PARAM_EXPR_NODE(pDefinedParams)->valuetype);

          pDefinedParams = NEXT_EXPR_NODE(pDefinedParams);
          pActParams     = NEXT_EXPR_NODE(pActParams);
	} //while
        /* check if defined == found number of parameters */
        if ( ((pDefinedParams != NULL) || (pActParams != NULL))
             && (res == 0) ) {
          //count the rest of the actParams (if any)
          while (pActParams != NULL) {
            actparam_count += 1;
	    pActParams = NEXT_EXPR_NODE(pActParams);
	  } //while
          pc_error(LINENO_NODE(node), COLNO_NODE(node), 
                   SYMBOL_NR_PARAMS(pSte), actparam_count);
          res = 1;
	} //if defined != found nr of params
      }
    } else {
      if (SYMBOL_NR_PARAMS(pSte) != 0) {
        pc_error(LINENO_NODE(node), COLNO_NODE(node), 
                 SYMBOL_NR_PARAMS(pSte), 0);
        res = 1;
      }
    } //if pActParams != NULL
    break;

  case e_listkeeper:
    DBUG_PRINT( "tc", ("e_listkeeper - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    node->valuetype = node->valuetype_expected;
    res = checkParams(LIST_EXPR_NODE(node), node->valuetype);
    break;

  case e_list:
    DBUG_PRINT( "tc", ("e_list - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    res = case_e_list(&checkParams, node);
    break;

  case e_prepend:
    DBUG_PRINT( "tc", ("e_prepend - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    /* set valuetype to the expected */
    DBUG_ASSERT(node->valuetype_expected > 0);
    node->valuetype = node->valuetype_expected;

    //is type really a MLList?
    listitem_type = getItemtypeOfList(node->valuetype);
    DBUG_ASSERT(listitem_type > -1); //-1 means type not found in symbol_table
    if (listitem_type == 0) {
      //type is not a list
      tc_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
               COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
               node->valuetype, list_type_index);
      printf("222-%p\n", node);
      res = 1;
    } else {
      /* only check the relavant children */
      res = checkParams(LEFT_OPERAND_EXPR_NODE(node), listitem_type);
      if (res == 0) {
        res = checkParams(RIGHT_OPERAND_EXPR_NODE(node), node->valuetype);
      }
    }
    break;

  case e_param_list:
    DBUG_PRINT( "tc", ("e_param_list - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    res = checkParams(PARAM_EXPR_NODE(node), parent_expects);
    if (res == 0) {
      res = checkParams(NEXT_EXPR_NODE(node), parent_expects);
    }
    break;

  default:
    DBUG_PRINT( "tc", ("defaultcase - %s (%d)", seekSymbolTypeindex(
                       symbol_table, node->valuetype_expected)->id, 
                       node->valuetype_expected) );
    printf_expr_node(node);
    tc_error(LINENO_NODE(node), COLNO_NODE(node), 0, 0);
    printf("333-%p\n", node);
    res = 1;
  } //switch

  //printf("leaving checkParams\n");
  DBUG_RETURN(res);
} //checkParams


/** @return 0 or 1 in case of an error
  */
static gint
paramsAction(GList *binding, Expr_node_ptr paramsroot, Expr_node_ptr ignored,
             MLType ignored2)
{
  DBUG_ENTER("paramsAction");
  act_binding = binding;//important for the correct work of checkParams
                        //has to be NULL while checking a constructor
  DBUG_RETURN( checkParams(paramsroot, 0) ); 
} //paramsAction


/** @return 0 or 1 in case of an error
  */
static gint
bodyAction(GList *binding, Expr_node_ptr ignored, Expr_node_ptr bodyroot, 
           MLType body_resulttype)
{
  gint res;
  MLType foundtype;

  DBUG_ENTER("bodyAction");

  act_binding = binding; //important for the correct work of checkParams
                         //has to be NULL while checking a constructor

  printf_symboltable(act_binding);
  foundtype = checkBody(bodyroot);
  if (foundtype == 0) {
    res = 1;
  } else {
    if (isFoundTypeCompatibleToExpected(body_resulttype, foundtype)) {
      res = 0;
    } else {
      tc_error(LINENO_NODE(bodyroot),
	       COLNO_NODE( bodyroot),
	       body_resulttype, foundtype);
      res = 1;
    }
  } //if foundtype == 0
  DBUG_RETURN(res); 
} //bodyAction


/** @return 0 or 1 in case of an error
  */
static gint
traverseParamsOrBody(GList* syntaxtree, NodeAction_funptr nodeAction)
{
  gint clauseCount, res = 0;
  Fun_node_ptr  pFun;
  GList *paramslist, *bindingslist, *bodylist, *tmp_st;

  DBUG_ENTER("traverseParamsOrBody");


  /*
   * run through all functions to check the parameters
   */
  tmp_st = syntaxtree;
  while ((res == 0) && (tmp_st != NULL)) {
    pFun = tmp_st->data;

    /* For datatypes there should only occur the case e_variable. The only 
     * thing to do is valuetype = valuetype_expected. */
    if (pFun->nodetype == e_constructor) {
      if (nodeAction == &paramsAction) {
	DBUG_PRINT("tc", ("Checking constructor '%s'", pFun->fun_id));
	//there is only one clause and no body
	paramslist = pFun->paramslist;
        //bodylist = g_list_prepand(NULL, NULL); //for safety
	clauseCount = 1;
	printf("clause %i", clauseCount);
	DBUG_PRINT("tc", ("clause %i", clauseCount));
	//check the pattern in parameter-lists
	res = (*nodeAction)(NULL, paramslist->data, NULL, 
                            pFun->resulttype);
	/*	
        act_binding  = NULL;/important for the correct work of checkParams
                            /has to be NULL while checking a constructor
			    res = checkParams(paramslist->data, 0);
        */
	//g_list_free(bodylist);
	//bodylist = NULL;
      }
    } //if e_constructor

    if (pFun->nodetype == e_userfun) {
	    printf("Checking fun '%s'\n", pFun->fun_id);
      DBUG_PRINT("tc", ("Checking fun '%s'", pFun->fun_id));
      //run through all clauses
      paramslist   = pFun->paramslist;
      bindingslist = pFun->bindingslist;
      bodylist     = pFun->bodylist;
      clauseCount = 1;
      while ((res == 0) && (paramslist != NULL ) && (bindingslist != NULL) && 
                           (bodylist != NULL)) {
        DBUG_PRINT("tc", ("clause %i", clauseCount));
	printf("clause %i\n", clauseCount);
        res = (*nodeAction)(bindingslist->data, paramslist->data, 
                            bodylist->data, pFun->resulttype);

        paramslist   = g_list_next(paramslist);
        bindingslist = g_list_next(bindingslist);
        bodylist     = g_list_next(bodylist);
        clauseCount++;
      } //while: paramslist && bodylist

      DBUG_ASSERT( ((paramslist == NULL ) && (bindingslist == NULL) && 
                    (bodylist == NULL)) || (res != 0) );
    } else {
      DBUG_ASSERT( (pFun->nodetype != e_userfun) || 
                   (pFun->nodetype != e_constructor) );
    }//if e_userfun

    tmp_st = g_list_next(tmp_st);
  } //while: tmp_st

  DBUG_RETURN(res);
} //traverseParamsOrBody


static void
seek_type_indices(void)
{
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("seek_type_indices");

  //seek the type_indices for basic types once only
  pSte = seekSymbolId(symbol_table, "unit");
  DBUG_ASSERT(pSte != NULL);
  unit_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "bool");
  DBUG_ASSERT(pSte != NULL);
  bool_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "int");
  DBUG_ASSERT(pSte != NULL);
  int_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "real");
  DBUG_ASSERT(pSte != NULL);
  real_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "string");
  DBUG_ASSERT(pSte != NULL);
  string_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "nil");
  DBUG_ASSERT(pSte != NULL);
  nil_type_index = pSte->type_index;

  pSte = seekSymbolId(symbol_table, "list");
  DBUG_ASSERT(pSte != NULL);
  list_type_index = pSte->type_index;

  DBUG_VOID_RETURN;
} //seek_type_indices


/** @return 0 or 1 in case of an error
  */
gint
typecheck_prg(GList* syntaxtree)
{
  gint res = 0;

  DBUG_ENTER("typecheck_prg");

  //seek the type_indices for basic types once only
  seek_type_indices();

  /*
   * run through all functions to check the parameters
   */
  res = traverseParamsOrBody(syntaxtree, &paramsAction);

  /*
   * run through all functions to check the bodies
   */
  if (res == 0) {
    res = traverseParamsOrBody(syntaxtree, &bodyAction);
  }

  DBUG_RETURN(res);
} //typecheck_prg


/** @return 0 or 1 in case of an error
  */
gint
typecheck_expr(Expr_node_ptr expr_root)
{
  gint res = 0;

  DBUG_ENTER("typecheck_prg");

  //seek the type_indices for basic types once only
  seek_type_indices();

  act_binding = NULL; /* important for the correct work of checkParams
		       * has to be NULL while checking a constructor
                       * and while checking a global expr */

  res = checkBody(expr_root);
  if (res == 0) {
    res = 1;
  } else {
    res = 0;
  }

  DBUG_RETURN(res);
} //typecheck_expr
